-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

local filtersWhiteList =
{ "Drivetrain"
, "Type"
, "Config Type"
, "Transmission"
, "Country"
, "Derby Class"
, "Performance Class"
, "Value"
, "Brand"
, "Body Style"
, "Source"
, "Weight"
, "Top Speed"
, "0-100 km/h"
, "0-60 mph"
, "Weight/Power"
, "Off-Road Score"
}

local range = {'Years'}

-- agregates only, attribute stays the same
local convertToRange = 
  { 'Value'
  , 'Weight'
  , 'Top Speed'
  , '0-100 km/h'
  , '0-60 mph'
  , 'Weight/Power'
  , "Off-Road Score"
  }

-- so th ui knows when to interpret the data as range
local finalRanges = {}
arrayConcat(finalRanges, range)
arrayConcat(finalRanges, convertToRange)

local displayInfo =
{ ranges =
  { all = finalRanges
  , real = range
  }
, units =
  { Weight = {type = 'weight', dec = 0}
  , ['Top Speed'] = {type = 'speed', dec = 0}
  , ['Torque'] = {type = 'torque', dec = 0}
  , ['Power'] = {type = 'power', dec = 0}
  , ['Weight/Power'] = {type = 'weightPower', dec = 2}
  }
, predefinedUnits =
  { ['0-60 mph'] = {unit = 's', type = 'speed', ifIs = 'mph', dec = 1}
  , ['0-120 mph'] = {unit = 's', type = 'speed', ifIs = 'mph', dec = 1}
  , ['0-200 mph'] = {unit = 's', type = 'speed', ifIs = 'mph', dec = 1}
  , ['60-120 mph'] = {unit = 's', type = 'speed', ifIs = 'mph', dec = 1}
  , ['60-0 mph'] = {unit = 'ft', type = 'distance', ifIs = 'mi', dec = 1}
  , ['0-100 km/h'] = {unit = 's', type = 'speed', ifIs = 'km/h', dec = 1}
  , ['0-200 km/h'] = {unit = 's', type = 'speed', ifIs = 'km/h', dec = 1}
  , ['0-300 km/h'] = {unit = 's', type = 'speed', ifIs = 'km/h', dec = 1}
  , ['100-200 km/h'] = {unit = 's', type = 'speed', ifIs = 'km/h', dec = 1}
  , ['100-0 km/h'] = {unit = 'm', type = 'distance', ifIs = 'km', dec = 1}
  }
, dontShowInDetails =
  { 'Type'
  , 'Config Type'
  }
, perfData =
  { '0-60 mph'
  , '0-120 mph'
  , '0-200 mph'
  , '60-120 mph'
  , '60-0 mph'
  , '0-100 km/h'
  , '0-200 km/h'
  , '0-300 km/h'
  , '100-200 km/h'
  , '100-0 km/h'
  , 'Top Speed'
  , 'Weight/Power'
  , 'Off-Road Score'
  }
, filterData = filtersWhiteList
}

-- TODO: Think about only operating on cache and not cache + local vairable in function

local files
local filesJBEAM
local filesPC

local showPcs = settings.getValue('showPcs')

local _modelNames

local cache = {}

local function _parseVehicleNameBackwardCompatibility(vehicleName)
    --print("_parseVehicleNameBackwardCompatibility: " .. tostring(vehicleName))

    -- get name.cs
    local nameCS = "vehicles/" .. vehicleName .. "/name.cs"
    local res = { configs = {} }
    local f = io.open(nameCS, "r")
    if f then
        for line in f:lines() do
            --print(line)
            local key, value = line:match("^%%(%w+)%s-=%s-\"(.+)\";")
            --print("key = " .. tostring(key))
            --print("value = " .. tostring(value))
            if key ~= nil and key == "vehicleName" and value ~= nil then
                res.Name = value
            end
        end
        f:close()
    end

    -- get .pc files and fix them up for the new system
    local pcfiles = FS:findFilesByRootPattern("game:vehicles/" .. vehicleName .. "/", "*.pc", true, true, false)
    for k,v in pairs(pcfiles) do
        local dir, filename, ext = path.split(v)
        if dir and filename and ext and ext == "pc" then
            local pcfn = filename:sub(1, #filename - 3)
            res.configs[pcfn] = { Configuration = pcfn}
        end
    end

    --really no name? :(
    if res.Name == nil then
        res.Name = vehicleName
    end

    -- bold assumption :|
    res.Type = "Car"

    return res
end




local function _fillAggregates(data, destination)
  for key, value in pairs(data) do
    if tableContains(range, key) then
      if not destination[key] then
        destination[key] = deepcopy(data[key])
      end

      destination[key].min = math.min(data[key].min, destination[key].min)
      destination[key].max = math.max(data[key].max, destination[key].max)
    elseif tableContains(convertToRange, key) then
      if type(data[key]) == 'number' then
        if not destination[key] then
          destination[key] = {min = data[key], max = data[key]}
        end
        destination[key].min = math.min(data[key], destination[key].min)
        destination[key].max = math.max(data[key], destination[key].max)
      end
    elseif tableContains(filtersWhiteList, key) then
      if not destination[key] then
        destination[key] = {}
      end

      destination[key][value] = true
    end
  end
end

local function _mergeAggregates(data, destination)
  for key, value in pairs(data) do
    if tableContains(range, key) or tableContains(convertToRange, key) then
      if not destination[key] then
        destination[key] = deepcopy(data[key])
      end

      destination[key].min = math.min(data[key].min, destination[key].min)
      destination[key].max = math.max(data[key].max, destination[key].max)
    elseif tableContains(filtersWhiteList, key) then
      if not destination[key] then
        destination[key] = deepcopy(data[key])
      else
        for key2, _ in pairs(value) do
          destination[key][key2] = true
        end
      end

    end
  end
end






-- gets all files related to vehicle info
local function _getFiles ()
  files       = FS:findFilesByRootPattern("game:/vehicles/", "*.json", true, true, true)
  filesJBEAM  = FS:findFilesByRootPattern("game:/vehicles/", "*.jbeam", true, true, true)
  filesPC     = FS:findFilesByRootPattern("game:/vehicles/", "*.pc", true, true, true)
end





-- returns all found model names
local function getModelNames ()
  if _modelNames then
    return _modelNames
  end

  -- name of a file or directory can have alphanumerics, hyphens and underscores.
  local modelRegex  = "vehicles/([%w|_|%-|%s]+)/$"

  if files == nil or filesJBEAM == nil then
    _getFiles()
  end

  local modelNames = {}

  -- Get the models. They are the directories one level under vehicles folder
  for _, path in ipairs(files) do
    if string.sub(path, 1, 3) == "DIR" then
      local model = string.match(path, modelRegex)
      if model then
        table.insert(modelNames, model)
      end
    end
  end

  -- fix bad vehicle mods without info.json :(
  for _, path in ipairs(filesJBEAM) do
    if string.sub(path, 1, 3) == "DIR" then
      local model = string.match(path, modelRegex)
      if model and model ~= "common" and not tableContains(modelNames, model) then
        table.insert(modelNames, model)
        log("E", "modmanager", "Vehicle '" .. model .. "' missing info.json")
      end
    end
  end
  _modelNames = modelNames
  return modelNames
end



-- Get the configuration-names for the specific model, i.e. the .json files inside model's folder
local function _getModelConfigNames (key, pcs, includePath)
  if key and tableContains(getModelNames(), key) then
    local configRegex = "vehicles/"..escape_magic(key).."/info_(.*)%.json"
    local configRegexPcs = "vehicles/"..escape_magic(key).."/(.*)%.pc"
    local configNames = {}
    local list = {}

    for _, path in ipairs(files) do
      local config = string.match(path, configRegex)

      if config then
        table.insert(configNames, config)
        list[path] = config
      end
    end

    if pcs then
      for _, path in ipairs(filesPC) do
        local config = string.match(path, configRegexPcs)

        if config and not tableContains(configNames, config) then
          table.insert(configNames, config)
          list[path] = config
        end
      end
    end

    if includePath then
      return list
    else
      return configNames
    end
  end

  return {}
end



local function _modelConfigsHelper (key, model)
  if key and tableContains(getModelNames(), key) then

    if cache[key].configs then
      return cache[key].configs
    end

    if not cache[key].configs then
      cache[key].configs = {}
    end

    local configNames = _getModelConfigNames(key, showPcs, true)

    local configs = {}

    for path, configName in pairs(configNames) do
      local readData = readJsonFile("game:vehicles/" .. key .. "/info_" .. configName .. ".json") or {}

      if isOfficialContent(FS:getFileRealPath(path)) then
        readData.Source = 'BeamNG - Official'
      elseif string.sub(path, -3) == '.pc' then
        readData.Source = 'Custom'
      else
        readData.Source = 'Mod'
      end

      local configData = readData

      if model.default_pc == nil then
        model.default_pc = configName
      end

      -- makes life easier
      configData.model_key = key
      configData.key = configName
      configData.aggregates = {}

      if not configData.Configuration then
        configData.Configuration = configName
      end
      configData.Name = model.Name .. " " .. configData.Configuration

      configData.preview = imageExistsDefault("vehicles/" .. key .. "/" .. configName .. ".png")

      if configData.default_color then
		if not string.find(configData.default_color,"^[0123456789.]+%s[0123456789.]+%s[0123456789.]+%s[0123456789.]+") then
			configData.default_colorname = configData.default_color
	        configData.default_color = model.colors[configData.default_colorname]
		end
      end

      if configData.Value then --if we have a value number
        configData.Value = tonumber(configData.Value) or configData.Value --make sure it's actually a NUMBER and not a string
      end

      configData.is_default_config = (configName == model.default_pc)

      if readData then
        _fillAggregates(readData, configData.aggregates)
      end

      configs[configName] = configData
    end

    return configs
  end

  return nil
end



-- get all info to one model
local function getModel (key)
  if key and tableContains(getModelNames(), key) then
    if cache[key] then
      return cache[key]
    end

    if not cache[key] then
      cache[key] = {}
    end


    local model = {}

    local data = readJsonFile("game:vehicles/"..key.."/info.json")

    local fixedVehicle = false
    if data == nil then
      data = _parseVehicleNameBackwardCompatibility(key)
      fixedVehicle = true
    end

    local realPath = FS:getFileRealPath("vehicles/"..key.."/info.json")

    -- Patch up old vehicles for new System
    local missingInfoConfigs = nil;
    if data.configs then
      missingInfoConfigs = data.configs
      for mConfigName, mConfig in pairs(missingInfoConfigs) do
        mConfig.is_default_config = false
        if not data.default_pc then
          data.default_pc = mConfigName
          mConfig.is_default_config = true
        end
        mConfig.aggregates = {}
        mConfig.Configuration = mConfigName
        mConfig.Name = data.Name .. ' ' .. mConfigName
        mConfig.key = mConfigName
        mConfig.model_key = key
        mConfig.preview = imageExistsDefault("vehicles/" .. key .. "/" .. mConfigName .. ".png")
      end
      data.configs = nil
    end

    if data then
      model = deepcopy(data)

      if not data.Type then
        model.Type = "Unknown"
        print("model", model, "has type \"Unknown\"")
      end

      model.aggregates = {} -- values for filtering
    end

    -- get preview if it exists
    model.preview = imageExistsDefault("vehicles/" .. key .. "/default.png")

    if model.default_color then
      model.default_colorname = model.default_color
      model.default_color = model.colors[model.default_colorname]
    else
      model.default_color = ""
      model.default_colorname = ""
    end

    model.key = key -- redundant but makes life easy

    cache[key].model = model

    cache[key].configs = missingInfoConfigs or _modelConfigsHelper(key, model)

    if cache[key].configs and tableSize(cache[key].configs) < 1 then
      cache[key].configs[key] = deepcopy(model)
      if (cache[key].configs[key].model_key == nil) then
        cache[key].configs[key].model_key = cache[key].configs[key].key
      end
      if cache[key].model.default_pc == nil then
        cache[key].model.default_pc = key
      end
      if isOfficialContent(realPath) then
        data.Source = 'BeamNG - Official'
      else
        data.Source = 'Mod'
      end
    end

    if fixedVehicle then
      data.Source = 'Mod'
    end

    if data then
      _fillAggregates(data, model.aggregates)
    end

    -- all configs should have the same aggregates as the base model
    -- the model should have all aggregates of the configs
    local aggHelper = {}
    for _, config in pairs(cache[key].configs) do
      _mergeAggregates(config.aggregates, aggHelper)
      _mergeAggregates(cache[key].model.aggregates, config.aggregates)
    end
    _mergeAggregates(aggHelper, cache[key].model.aggregates)

    return cache[key]
  end

  return nil
end


-- returns the key of the current vehicle (of player one)
-- one could also use: be:getPlayerVehicle(0):getJBeamFilename()
local function getCurrentVehicleDetails ()
  local res = readJsonData(TorqueScript.call('getVehicle'))

  res.key = res.model
  res.model = nil

  local model = getModel(res.key)

  local config = {}

  if res.configuration then
    res.key = string.match(res.configuration, "vehicles/".. res.key .."/info_(.*).json")
    res.pc_file = res.configuration
    res.configuration = nil

    config = model.configs[res.key]
  end

  return {current = res, model = model.model, configs = config}
end




local function createFilters (list)
  local filter = {}

  if list then
    for _, value in pairs(list) do
      for propName, propVal in pairs(value.aggregates) do

        if tableContains(finalRanges, propName) then
          if filter[propName] then
            filter[propName].min = math.min(value.aggregates[propName].min, filter[propName].min)
            filter[propName].max = math.max(value.aggregates[propName].max, filter[propName].max)
          else
            filter[propName] = deepcopy(value.aggregates[propName])
          end
        else
          if not filter[propName] then
            filter[propName] = {}
          end
          for key,_ in pairs(propVal) do
            filter[propName][key .. ''] = true
          end
        end
      end
    end
  end

  return filter
end


-- get the list of all available models
local function getModelList (array)
  local models = {}
  local modelNames = getModelNames()

  for _, value in pairs(modelNames) do
    local model = getModel(value)
    if array then
      table.insert(models, model.model)
    else
      models[model.model.key] = model.model
    end
  end

  return {models = models, filters = createFilters(models), displayInfo = displayInfo}
end

-- get the list of all available configurations
local function getConfigList (array)
  local configList = {}
  local modelNames = getModelNames()

  for _, value in pairs(modelNames) do
    local model = getModel(value)
    -- dump(model.configs)
    if model.configs and tableSize(model.configs) > 0 then
      for _, config in pairs(model.configs) do
        if array then
          table.insert(configList, config)
        else
          -- dump(config)
          configList[config.model_key .. '_' .. config.key] = config
        end
      end
    end
  end

  return {configs = configList, filters = createFilters(configList), displayInfo = displayInfo}
end


-- get the list of all available vehicles
-- mainly thought for ui
local function sendList ()
  local models = getModelList(true)
  local configs = getConfigList(true, true)
  guihooks.trigger('sendVehicleList', {models = models.models, configs = configs.configs, filters = models.filters, displayInfo = displayInfo})
end




local function spawnFilldefaults (key, opt)
  local model = getModel(key)
  -- dump(model)
  if not opt then
    opt = {}
  end


  if opt.color and type(opt.color) == 'string' and model.model.colors[opt.color] then
    opt.color = model.model.colors[opt.color]
  end

  if not opt.color and opt.config then
    local config = model.configs[opt.config]
    if config then
      opt.color = config.default_color
    end
  end

  if not opt.color then
    opt.color = model.model.default_color
  end


  if not opt.config then
    opt.config = 'vehicles/' .. key .. '/' .. model.model.default_pc .. '.pc'
  else
    opt.config = 'vehicles/' .. key .. '/' .. opt.config .. '.pc'
  end

  return opt
end

-- opt: {config, color, pos}
-- config: key
-- color can be a key from the colors list or an rgb string with spaces as seperators i.e "5 5 5"
-- pos should be a string with xyz coordinates as string with spaces as seperators i.e "5 5 5"
local function spawnNewVehicle (key, opt)
  local firstVehicle = (be:getObjectCount() == 0)
  local options = spawnFilldefaults(key, opt)

  if options.pos then
    TorqueScript.call('spawnSpecificVehicleAtPos', key, options.config, options.pos, options.color)
  else
    TorqueScript.call('spawnVehicleInplace', key, options.config, options.color)
  end

  if firstVehicle then
    Lua:setControlVehicle(be:getObject(0), 0)
  end

  if be:getObjectCount() > 1 then
    ui_message("Press TAB to switch vehicle", 10, "spawn")
  end
end

-- opt: {config, color, reload}
-- config: key
-- color can be a key from the colors list or an rgb string with spaces as seperators i.e "5 5 5"
local function replaceVehicle (key, opt)
  -- when no vehicle is spawned, spawn a new one instead
  if be:getObjectCount() == 0 then
    spawnNewVehicle(key, opt)
    return
  end
  -- dump(key, opt)
  local options = spawnFilldefaults(key, opt)

  if options.reload ~= nil then
    TorqueScript.call('chooseVehicle', key, options.config, options.color, options.reload)
  else
    TorqueScript.call('chooseVehicle', key, options.config, options.color)
  end
end

local function removeCurrent()
  local veh = Lua:getControlVehicle()
  if veh then
    veh:delete()
  end
  if be:getObjectCount() > 0 then
    -- enter first vehicle instead
    Lua:setControlVehicle(be:getObject(0), 0)
  else
    -- switch to free camera
    TorqueScript.call('LocalClientConnection.setControlObject();')
    ui_message("No vehicles available, switching to free camera", 15, "camera")
  end
end

local function removeAllExceptCurrent()
  local lveh = Lua:getControlVehicle()
  local vid = lveh:getID()
  while be:getObjectCount() > 1 do
    for i = 0, be:getObjectCount() do
      local veh = be:getObject(i)
      if veh and veh:getID() ~= vid then
        veh:delete()
      end
    end
  end
end

local function removeAll()
  while be:getObjectCount() > 0 do
    local veh = be:getObject(0)
    if veh then
      veh:delete()
    end
  end
  -- switch to free camera
  TorqueScript.call('LocalClientConnection.setControlObject(LocalClientConnection.camera);')
  ui_message("No vehicles available, switching to free camera", 15, "camera")
end

local function onFileChanged(filename, type)
  -- TODO: think about only deleting the according data from the cache
  if string.find(filename, 'vehicles/') == 1 or string.find(filename, 'mods/') == 1 then
    files = nil
    filesJBEAM = nil
    cache = {}
    _modelNames = nil
    sendList()

    -- print('Throw away vehicles.lua\'s cache')
  end
end

local function onSettingsChanged()
  if showPcs ~= settings.getValue('showPcs') then
    files = nil
    filesJBEAM = nil
    cache = {}
    _modelNames = nil
    showPcs = settings.getValue('showPcs')
    sendList()

    -- print('Throw away vehicles.lua\'s cache')
  end
end

local function numConfigOfModel (model)
  return tableSize(_getModelConfigNames(model, true))
end

--public interface
M.getCurrentVehicle = getCurrentVehicleDetails

M.getModel = getModel
M.requestList = sendList
M.getModelList = getModelList
M.getConfigList = getConfigList

M.replaceVehicle = replaceVehicle
M.spawnNewVehicle = spawnNewVehicle
M.removeCurrent = removeCurrent
M.removeAll = removeAll
M.removeAllExceptCurrent = removeAllExceptCurrent

M.numConfigOfModel = numConfigOfModel

-- not sure if needed (currently only used by jan to determine which mod was deactivated)
M.getModelNames = getModelNames

-- used to delete the cached data
M.onFileChanged = onFileChanged
M.onSettingsChanged = onSettingsChanged


return M
